# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Dependency/topological sort implementation.

"""

from suds import *

from logging import getLogger
log = getLogger(__name__)


def dependency_sort(dependency_tree):
    """
    Sorts items 'dependencies first' in a given dependency tree.

    A dependency tree is a dictionary mapping an object to a collection its
    dependency objects.

    Result is a properly sorted list of items, where each item is a 2-tuple
    containing an object and its dependency list, as given in the input
    dependency tree.

    If B is directly or indirectly dependent on A and they are not both a part
    of the same dependency cycle (i.e. then A is neither directly nor
    indirectly dependent on B) then A needs to come before B.

    If A and B are a part of the same dependency cycle, i.e. if they are both
    directly or indirectly dependent on each other, then it does not matter
    which comes first.

    Any entries found listed as dependencies, but that do not have their own
    dependencies listed as well, are logged & ignored.

    @return: The sorted items.
    @rtype: list

    """
    sorted = []
    processed = set()
    for key, deps in dependency_tree.items():
        _sort_r(sorted, processed, key, deps, dependency_tree)
    return sorted


def _sort_r(sorted, processed, key, deps, dependency_tree):
    """Recursive topological sort implementation."""
    if key in processed:
        return
    processed.add(key)
    for dep_key in deps:
        dep_deps = dependency_tree.get(dep_key)
        if dep_deps is None:
            log.debug('"%s" not found, skipped', Repr(dep_key))
            continue
        _sort_r(sorted, processed, dep_key, dep_deps, dependency_tree)
    sorted.append((key, deps))



# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Provides modules containing classes to support marshalling to XML.

"""

from suds.sudsobject import Object


class Content(Object):
    """
    Marshaller content.

    @ivar tag: The content tag.
    @type tag: str
    @ivar value: The content's value.
    @type value: I{any}

    """

    extensions = []

    def __init__(self, tag=None, value=None, **kwargs):
        """
        @param tag: The content tag.
        @type tag: str
        @param value: The content's value.
        @type value: I{any}

        """
        Object.__init__(self)
        self.tag = tag
        self.value = value
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getattr__(self, name):
        try:
            return self.__dict__[name]
        except KeyError:
            pass
        if name in self.extensions:
            value = None
            setattr(self, name, value)
            return value
        raise AttributeError("Content has no attribute %s" % (name,))



# This program is free software; you can redistribute it and/or modify
# it under the terms of the (LGPL) GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library Lesser General Public License for more details at
# ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
The I{builder} module provides an wsdl/xsd defined types factory
"""

from suds import *
from suds.sudsobject import Factory


class Builder:
    """ Builder used to construct an object for types defined in the schema """

    def __init__(self, resolver):
        """
        @param resolver: A schema object name resolver.
        @type resolver: L{resolver.Resolver}
        """
        self.resolver = resolver

    def build(self, name):
        """ build a an object for the specified typename as defined in the schema """
        if isinstance(name, str):
            type = self.resolver.find(name)
            if type is None:
                raise TypeNotFound(name)
        else:
            type = name
        cls = type.name
        if type.mixed():
            data = Factory.property(cls)
        else:
            data = Factory.object(cls)
        resolved = type.resolve()
        md = data.__metadata__
        md.sxtype = resolved
        md.ordering = self.ordering(resolved)
        history = []
        self.add_attributes(data, resolved)
        for child, ancestry in type.children():
            if self.skip_child(child, ancestry):
                continue

            self.process(data, child, history[:])
        return data

    def process(self, data, type, history):
        """ process the specified type then process its children """
        if type in history:
            return
        if type.enum():
            return
        history.append(type)
        resolved = type.resolve()
        value = None



        if type.multi_occurrence():
            value = []
        else:
            if len(resolved) > 0:
                if resolved.mixed():
                    value = Factory.property(resolved.name)
                    md = value.__metadata__
                    md.sxtype = resolved
                else:
                    value = Factory.object(resolved.name)
                    md = value.__metadata__
                    md.sxtype = resolved
                    md.ordering = self.ordering(resolved)

        setattr(data, type.name, None if self.skip_value(type) else value)
        if value is not None:
            data = value
        if not isinstance(data, list):
            self.add_attributes(data, resolved)
            for child, ancestry in resolved.children():
                if self.skip_child(child, ancestry):
                    continue
                self.process(data, child, history[:])

    def add_attributes(self, data, type):
        """ add required attributes """
        for attr, ancestry in type.attributes():
            name = '_%s' % attr.name
            value = attr.get_default()
            setattr(data, name, value)

    def skip_child(self, child, ancestry):
        """ get whether or not to skip the specified child """
        if child.any(): return True
        for x in ancestry:
            if x.choice():
                return True
        return False

    def skip_value(self, type):
        """ whether or not to skip setting the value """
        return type.optional() and not type.multi_occurrence()

    def ordering(self, type):
        """ get the ordering """
        result = []
        for child, ancestry in type.resolve():
            name = child.name
            if child.name is None:
                continue
            if child.isattr():
                name = '_%s' % child.name
            result.append(name)
        return result



# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Provides XML I{attribute} classes.

"""

from suds import UnicodeMixin
from suds.sax import splitPrefix, Namespace
from suds.sax.text import Text


class Attribute(UnicodeMixin):
    """
    An XML attribute object.

    @ivar parent: The node containing this attribute.
    @type parent: L{element.Element}
    @ivar prefix: The I{optional} namespace prefix.
    @type prefix: basestring
    @ivar name: The I{unqualified} attribute name.
    @type name: basestring
    @ivar value: The attribute's value.
    @type value: basestring

    """

    def __init__(self, name, value=None):
        """
        @param name: The attribute's name with I{optional} namespace prefix.
        @type name: basestring
        @param value: The attribute's value.
        @type value: basestring

        """
        self.parent = None
        self.prefix, self.name = splitPrefix(name)
        self.setValue(value)

    def clone(self, parent=None):
        """
        Clone this object.

        @param parent: The parent for the clone.
        @type parent: L{element.Element}
        @return: A copy of this object assigned to the new parent.
        @rtype: L{Attribute}

        """
        a = Attribute(self.qname(), self.value)
        a.parent = parent
        return a

    def qname(self):
        """
        Get this attribute's B{fully} qualified name.

        @return: The fully qualified name.
        @rtype: basestring

        """
        if self.prefix is None:
            return self.name
        return ":".join((self.prefix, self.name))

    def setValue(self, value):
        """
        Set the attribute's value.

        @param value: The new value (may be None).
        @type value: basestring
        @return: self
        @rtype: L{Attribute}

        """
        if isinstance(value, Text):
            self.value = value
        else:
            self.value = Text(value)
        return self

    def getValue(self, default=Text("")):
        """
        Get the attributes value with optional default.

        @param default: An optional value to return when the attribute's value
            has not been set.
        @type default: basestring
        @return: The attribute's value, or I{default}.
        @rtype: L{Text}

        """
        return self.value or default

    def hasText(self):
        """
        Get whether the attribute has a non-empty I{text} string value.

        @return: True when has I{text}.
        @rtype: boolean

        """
        return bool(self.value)

    def namespace(self):
        """
        Get the attribute's namespace. This may either be the namespace defined
        by an optional prefix, or the default namespace.

        @return: The attribute's namespace.
        @rtype: (I{prefix}, I{name})

        """
        if self.prefix is None:
            return Namespace.default
        return self.resolvePrefix(self.prefix)

    def resolvePrefix(self, prefix):
        """
        Resolve the specified prefix to a known namespace.

        @param prefix: A declared prefix.
        @type prefix: basestring
        @return: The namespace mapped to I{prefix}.
        @rtype: (I{prefix}, I{name})

        """
        if self.parent is None:
            return Namespace.default
        return self.parent.resolvePrefix(prefix)

    def match(self, name=None, ns=None):
        """
        Match by (optional) name and/or (optional) namespace.

        @param name: The optional attribute tag name.
        @type name: str
        @param ns: An optional namespace.
        @type ns: (I{prefix}, I{name})
        @return: True if matched.
        @rtype: boolean

        """
        byname = name is None or (self.name == name)
        byns = ns is None or (self.namespace()[1] == ns[1])
        return byname and byns

    def __eq__(self, rhs):
        """Equals operator."""
        return (isinstance(rhs, Attribute) and self.prefix == rhs.name and
            self.name == rhs.name)

    def __repr__(self):
        """Programmer friendly string representation."""
        return "attr (prefix=%s, name=%s, value=(%s))" % (self.prefix,
            self.name, self.value)

    def __unicode__(self):
        """XML string representation."""
        return '%s="%s"' % (self.qname(), self.value and self.value.escape())



# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Provides literal I{marshaller} classes.

"""

from suds import *
from suds.mx import *
from suds.mx.core import Core
from suds.mx.typer import Typer
from suds.resolver import Frame, GraphResolver
from suds.sax.element import Element
from suds.sudsobject import Factory

from logging import getLogger
log = getLogger(__name__)


# add typed extensions
Content.extensions.append("type")  # The expected xsd type
Content.extensions.append("real")  # The 'true' XSD type
Content.extensions.append("ancestry")  # The 'type' ancestry


class Typed(Core):
    """
    A I{typed} marshaller.

    This marshaller is semi-typed as needed to support both I{document/literal}
    and I{rpc/literal} SOAP message styles.

    @ivar schema: An XSD schema.
    @type schema: L{xsd.schema.Schema}
    @ivar resolver: A schema type resolver.
    @type resolver: L{GraphResolver}

    """

    def __init__(self, schema, xstq=True):
        """
        @param schema: A schema object
        @type schema: L{xsd.schema.Schema}
        @param xstq: The B{x}ml B{s}chema B{t}ype B{q}ualified flag indicates
            that the I{xsi:type} attribute values should be qualified by
            namespace.
        @type xstq: bool

        """
        Core.__init__(self)
        self.schema = schema
        self.xstq = xstq
        self.resolver = GraphResolver(self.schema)

    def reset(self):
        self.resolver.reset()

    def start(self, content):
        """
        Start marshalling the 'content' by ensuring that both the 'content'
        _and_ the resolver are primed with the XSD type information. The
        'content' value is both translated and sorted based on the XSD type.
        Only values that are objects have their attributes sorted.

        """
        log.debug("starting content:\n%s", content)
        if content.type is None:
            name = content.tag
            if name.startswith("_"):
                name = "@" + name[1:]
            content.type = self.resolver.find(name, content.value)
            if content.type is None:
                raise TypeNotFound(content.tag)
        else:
            known = None
            if isinstance(content.value, Object):
                known = self.resolver.known(content.value)
                if known is None:
                    log.debug("object %s has no type information",
                        content.value)
                    known = content.type
            frame = Frame(content.type, resolved=known)
            self.resolver.push(frame)
        frame = self.resolver.top()
        content.real = frame.resolved
        content.ancestry = frame.ancestry
        self.translate(content)
        self.sort(content)
        if self.skip(content):
            log.debug("skipping (optional) content:\n%s", content)
            self.resolver.pop()
            return False
        return True

    def suspend(self, content):
        """
        Suspend to process list content.

        Primarily, this involves popping the 'list' content off the resolver's
        stack so its list items can be marshalled.

        """
        self.resolver.pop()

    def resume(self, content):
        """
        Resume processing list content.

        To do this, we really need to simply push the 'list' content back onto
        the resolver stack.

        """
        self.resolver.push(Frame(content.type))

    def end(self, parent, content):
        """
        End processing the content.

        Make sure the content ending matches the top of the resolver stack
        since for list processing we play games with the resolver stack.

        """
        log.debug("ending content:\n%s", content)
        current = self.resolver.top().type
        if current != content.type:
            raise Exception("content (end) mismatch: top=(%s) cont=(%s)" % (
                current, content))
        self.resolver.pop()

    def node(self, content):
        """
        Create an XML node.

        The XML node is namespace qualified as defined by the corresponding
        schema element.

        """
        ns = content.type.namespace()
        if content.type.form_qualified:
            node = Element(content.tag, ns=ns)
            if ns[0]:
                node.addPrefix(ns[0], ns[1])
        else:
            node = Element(content.tag)
        self.encode(node, content)
        log.debug("created - node:\n%s", node)
        return node

    def setnil(self, node, content):
        """
        Set the 'node' nil only if the XSD type specifies that it is permitted.

        """
        if content.type.nillable:
            node.setnil()

    def setdefault(self, node, content):
        """Set the node to the default value specified by the XSD type."""
        default = content.type.default
        if default is not None:
            node.setText(default)
        return default

    def optional(self, content):
        if content.type.optional():
            return True
        for a in content.ancestry:
            if a.optional():
                return True
        return False

    def encode(self, node, content):
        """
        Add (SOAP) encoding information if needed.

        The encoding information is added only if the resolved type is derived
        by extension. Furthermore, the xsi:type value is qualified by namespace
        only if the content (tag) and referenced type are in different
        namespaces.

        """
        if content.type.any():
            return
        if not content.real.extension():
            return
        if content.type.resolve() == content.real:
            return
        ns = None
        name = content.real.name
        if self.xstq:
            ns = content.real.namespace("ns1")
        Typer.manual(node, name, ns)

    def skip(self, content):
        """
        Get whether to skip this I{content}.

        Should be skipped when the content is optional and value is either None
        or an empty list.

        @param content: Content to skip.
        @type content: L{Object}
        @return: True if content is to be skipped.
        @rtype: bool

        """
        if self.optional(content):
            v = content.value
            if v is None:
                return True
            if isinstance(v, (list, tuple)) and not v:
                return True
        return False

    def optional(self, content):
        if content.type.optional():
            return True
        for a in content.ancestry:
            if a.optional():
                return True
        return False

    def translate(self, content):
        """
        Translate using the XSD type information.

        Python I{dict} is translated to a suds object. Most importantly,
        primitive values are translated from python to XML types using the XSD
        type.

        @param content: Content to translate.
        @type content: L{Object}
        @return: self
        @rtype: L{Typed}

        """
        v = content.value
        if v is None:
            return
        if isinstance(v, dict):
            cls = content.real.name
            content.value = Factory.object(cls, v)
            md = content.value.__metadata__
            md.sxtype = content.type
            return
        v = content.real.translate(v, False)
        content.value = v
        return self

    def sort(self, content):
        """
        Sort suds object attributes.

        The attributes are sorted based on the ordering defined in the XSD type
        information.

        @param content: Content to sort.
        @type content: L{Object}
        @return: self
        @rtype: L{Typed}

        """
        v = content.value
        if isinstance(v, Object):
            md = v.__metadata__
            md.ordering = self.ordering(content.real)
        return self

    def ordering(self, type):
        """
        Attribute ordering defined in the specified XSD type information.

        @param type: XSD type object.
        @type type: L{SchemaObject}
        @return: An ordered list of attribute names.
        @rtype: list

        """
        result = []
        for child, ancestry in type.resolve():
            name = child.name
            if child.name is None:
                continue
            if child.isattr():
                name = "_%s" % (child.name,)
            result.append(name)
        return result


class Literal(Typed):
    """
    A I{literal} marshaller.

    This marshaller is semi-typed as needed to support both I{document/literal}
    and I{rpc/literal} soap message styles.

    """
    pass



# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Classes for the (WS) SOAP I{document/literal} binding.

"""

from suds import *
from suds.argparser import parse_args
from suds.bindings.binding import Binding
from suds.sax.element import Element


class Document(Binding):
    """
    The document/literal style. Literal is the only (@use) supported since
    document/encoded is pretty much dead.

    Although the SOAP specification supports multiple documents within the SOAP
    <body/>, it is very uncommon. As such, suds library supports presenting an
    I{RPC} view of service methods defined with only a single document
    parameter. To support the complete specification, service methods defined
    with multiple documents (multiple message parts), are still presented using
    a full I{document} view.

    More detailed description:

    An interface is considered I{wrapped} if:
      - There is exactly one message part in that interface.
      - The message part resolves to an element of a non-builtin type.
    Otherwise it is considered I{bare}.

    I{Bare} interface is interpreted directly as specified in the WSDL schema,
    with each message part represented by a single parameter in the suds
    library web service operation proxy interface (input or output).

    I{Wrapped} interface is interpreted without the external wrapping document
    structure, with each of its contained elements passed through suds
    library's web service operation proxy interface (input or output)
    individually instead of as a single I{document} object.

    """
    def bodycontent(self, method, args, kwargs):
        wrapped = method.soap.input.body.wrapped
        if wrapped:
            pts = self.bodypart_types(method)
            root = self.document(pts[0])
        else:
            root = []

        def add_param(param_name, param_type, in_choice_context, value):
            """
            Construct request data for the given input parameter.

            Called by our argument parser for every input parameter, in order.

            A parameter's type is identified by its corresponding XSD schema
            element.

            """
            # Do not construct request data for undefined input parameters
            # defined inside a choice order indicator. An empty choice
            # parameter can still be included in the constructed request by
            # explicitly providing an empty string value for it.
            #TODO: This functionality might be better placed inside the
            # mkparam() function but to do that we would first need to better
            # understand how different Binding subclasses in suds work and how
            # they would be affected by this change.
            if in_choice_context and value is None:
                return

            # Construct request data for the current input parameter.
            pdef = (param_name, param_type)
            p = self.mkparam(method, pdef, value)
            if p is None:
                return
            if not wrapped:
                ns = param_type.namespace("ns0")
                p.setPrefix(ns[0], ns[1])
            root.append(p)

        parse_args(method.name, self.param_defs(method), args, kwargs,
            add_param, self.options().extraArgumentErrors)

        return root

    def replycontent(self, method, body):
        if method.soap.output.body.wrapped:
            return body[0].children
        return body.children

    def document(self, wrapper):
        """
        Get the document root. For I{document/literal}, this is the name of the
        wrapper element qualified by the schema's target namespace.

        @param wrapper: The method name.
        @type wrapper: L{xsd.sxbase.SchemaObject}
        @return: A root element.
        @rtype: L{Element}

        """
        tag = wrapper[1].name
        ns = wrapper[1].namespace("ns0")
        return Element(tag, ns=ns)

    def mkparam(self, method, pdef, object):
        """
        Expand list parameters into individual parameters each with the type
        information. This is because in document arrays are simply
        multi-occurrence elements.

        """
        if isinstance(object, (list, tuple)):
            return [self.mkparam(method, pdef, item) for item in object]
        return super(Document, self).mkparam(method, pdef, object)

    def param_defs(self, method):
        """Get parameter definitions for document literal."""
        pts = self.bodypart_types(method)
        if not method.soap.input.body.wrapped:
            return pts
        pt = pts[0][1].resolve()
        return [(c.name, c, a) for c, a in pt if not c.isattr()]

    def returned_types(self, method):
        rts = super(Document, self).returned_types(method)
        if not method.soap.output.body.wrapped:
            return rts
        return [child for child, ancestry in rts[0].resolve(nobuiltin=True)]



# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Provides a collection of suds objects primarily used for highly dynamic
interactions with WSDL/XSD defined types.

"""

from suds import *

from logging import getLogger
log = getLogger(__name__)


def items(sobject):
    """
    Extract the I{items} from a suds object.

    Much like the items() method works on I{dict}.

    @param sobject: A suds object
    @type sobject: L{Object}
    @return: A list of items contained in I{sobject}.
    @rtype: [(key, value),...]

    """
    for item in sobject:
        yield item


def asdict(sobject, recursive=False):
    """
    Convert a sudsobject into a dictionary.

    @param sobject: A suds object
    @type sobject: L{Object}
    @return: A python dictionary containing the items contained in I{sobject}.
    @rtype: dict

    """
    if recursive:
        return recursive_asdict(sobject)
    return dict(items(sobject))

def recursive_asdict(suds_object):
  """Convert a suds object to a dictionary.

  Args:
    suds_object: A suds object

  Returns:
    A python dictionary containing the items contained in the suds object.
  """

  # Create an empty dictionary.
  output_dict = {}

  # Iterate over the items in the suds object.
  for key, value in asdict(suds_object).items():

    # If the value is a suds object, recursively convert it to a dictionary.
    if hasattr(value, "__keylist__"):
      output_dict[key] = recursive_asdict(value)

    # If the value is a list, recursively convert each item in the list to a dictionary.
    elif isinstance(value, list):
      output_dict[key] = []
      for item in value:
        if hasattr(item, "__keylist__"):
          output_dict[key].append(recursive_asdict(item))
        else:
          output_dict[key].append(item)

    # Otherwise, just add the value to the dictionary.
    else:
      output_dict[key] = value

  # Return the dictionary.
  return output_dict


def merge(a, b):
    """
    Merge all attributes and metadata from I{a} to I{b}.

    @param a: A I{source} object
    @type a: L{Object}
    @param b: A I{destination} object
    @type b: L{Object}

    """
    for item in a:
        setattr(b, item[0], item[1])
        b.__metadata__ = b.__metadata__
    return b

def footprint(sobject):
    """
    Get the I{virtual footprint} of the object.

    This is really a count of all the significant value attributes in the
    branch.

    @param sobject: A suds object.
    @type sobject: L{Object}
    @return: The branch footprint.
    @rtype: int

    """
    n = 0
    for a in sobject.__keylist__:
        v = getattr(sobject, a)
        if v is None:
            continue
        if isinstance(v, Object):
            n += footprint(v)
            continue
        if hasattr(v, "__len__"):
            if len(v):
                n += 1
            continue
        n += 1
    return n


class Factory:

    cache = {}

    @classmethod
    def subclass(cls, name, bases, dict={}):
        if not isinstance(bases, tuple):
            bases = (bases,)
        # name is of type unicode in python 2 -> not accepted by type()
        name = str(name)
        key = ".".join((name, str(bases)))
        subclass = cls.cache.get(key)
        if subclass is None:
            subclass = type(name, bases, dict)
            cls.cache[key] = subclass
        return subclass

    @classmethod
    def object(cls, classname=None, dict={}):
        if classname is not None:
            subclass = cls.subclass(classname, Object)
            inst = subclass()
        else:
            inst = Object()
        for a in list(dict.items()):
            setattr(inst, a[0], a[1])
        return inst

    @classmethod
    def metadata(cls):
        return Metadata()

    @classmethod
    def property(cls, name, value=None):
        subclass = cls.subclass(name, Property)
        return subclass(value)


class Object(UnicodeMixin):

    def __init__(self):
        self.__keylist__ = []
        self.__printer__ = Printer()
        self.__metadata__ = Metadata()

    def __setattr__(self, name, value):
        builtin = name.startswith("__") and name.endswith("__")
        if not builtin and name not in self.__keylist__:
            self.__keylist__.append(name)
        self.__dict__[name] = value

    def __delattr__(self, name):
        try:
            del self.__dict__[name]
            builtin = name.startswith("__") and name.endswith("__")
            if not builtin:
                self.__keylist__.remove(name)
        except Exception:
            cls = self.__class__.__name__
            raise AttributeError("%s has no attribute '%s'" % (cls, name))

    def __getitem__(self, name):
        if isinstance(name, int):
            name = self.__keylist__[int(name)]
        return getattr(self, name)

    def __setitem__(self, name, value):
        setattr(self, name, value)

    def __iter__(self):
        return Iter(self)

    def __len__(self):
        return len(self.__keylist__)

    def __contains__(self, name):
        return name in self.__keylist__

    def __repr__(self):
        return str(self)

    def __unicode__(self):
        return self.__printer__.tostr(self)


class Iter:

    def __init__(self, sobject):
        self.sobject = sobject
        self.keylist = self.__keylist(sobject)
        self.index = 0

    def __next__(self):
        keylist = self.keylist
        nkeys = len(self.keylist)
        while self.index < nkeys:
            k = keylist[self.index]
            self.index += 1
            if hasattr(self.sobject, k):
                v = getattr(self.sobject, k)
                return (k, v)
        raise StopIteration()

    def __keylist(self, sobject):
        keylist = sobject.__keylist__
        try:
            keyset = set(keylist)
            ordering = sobject.__metadata__.ordering
            ordered = set(ordering)
            if not ordered.issuperset(keyset):
                log.debug("%s must be superset of %s, ordering ignored",
                    keylist, ordering)
                raise KeyError()
            return ordering
        except Exception:
            return keylist

    def __iter__(self):
        return self


class Metadata(Object):
    def __init__(self):
        self.__keylist__ = []
        self.__printer__ = Printer()


class Facade(Object):
    def __init__(self, name):
        Object.__init__(self)
        md = self.__metadata__
        md.facade = name


class Property(Object):

    def __init__(self, value):
        Object.__init__(self)
        self.value = value

    def items(self):
        for item in self:
            if item[0] != "value":
                yield item

    def get(self):
        return self.value

    def set(self, value):
        self.value = value
        return self


class Printer:
    """Pretty printing of a Object object."""

    @classmethod
    def indent(cls, n):
        return "%*s" % (n * 3, " ")

    def tostr(self, object, indent=-2):
        """Get s string representation of object."""
        history = []
        return self.process(object, history, indent)

    def process(self, object, h, n=0, nl=False):
        """Print object using the specified indent (n) and newline (nl)."""
        if object is None:
            return "None"
        if isinstance(object, Object):
            if len(object) == 0:
                return "<empty>"
            return self.print_object(object, h, n + 2, nl)
        if isinstance(object, dict):
            if len(object) == 0:
                return "<empty>"
            return self.print_dictionary(object, h, n + 2, nl)
        if isinstance(object, (list, tuple)):
            if len(object) == 0:
                return "<empty>"
            return self.print_collection(object, h, n + 2)
        if isinstance(object, str):
            return '"%s"' % (tostr(object),)
        return "%s" % (tostr(object),)

    def print_object(self, d, h, n, nl=False):
        """Print complex using the specified indent (n) and newline (nl)."""
        s = []
        cls = d.__class__
        if d in h:
            s.append("(")
            s.append(cls.__name__)
            s.append(")")
            s.append("...")
            return "".join(s)
        h.append(d)
        if nl:
            s.append("\n")
            s.append(self.indent(n))
        if cls != Object:
            s.append("(")
            if isinstance(d, Facade):
                s.append(d.__metadata__.facade)
            else:
                s.append(cls.__name__)
            s.append(")")
        s.append("{")
        for item in d:
            if self.exclude(d, item):
                continue
            item = self.unwrap(d, item)
            s.append("\n")
            s.append(self.indent(n+1))
            if isinstance(item[1], (list,tuple)):
                s.append(item[0])
                s.append("[]")
            else:
                s.append(item[0])
            s.append(" = ")
            s.append(self.process(item[1], h, n, True))
        s.append("\n")
        s.append(self.indent(n))
        s.append("}")
        h.pop()
        return "".join(s)

    def print_dictionary(self, d, h, n, nl=False):
        """Print complex using the specified indent (n) and newline (nl)."""
        if d in h:
            return "{}..."
        h.append(d)
        s = []
        if nl:
            s.append("\n")
            s.append(self.indent(n))
        s.append("{")
        for item in list(d.items()):
            s.append("\n")
            s.append(self.indent(n+1))
            if isinstance(item[1], (list,tuple)):
                s.append(tostr(item[0]))
                s.append("[]")
            else:
                s.append(tostr(item[0]))
            s.append(" = ")
            s.append(self.process(item[1], h, n, True))
        s.append("\n")
        s.append(self.indent(n))
        s.append("}")
        h.pop()
        return "".join(s)

    def print_collection(self, c, h, n):
        """Print collection using the specified indent (n) and newline (nl)."""
        if c in h:
            return "[]..."
        h.append(c)
        s = []
        for item in c:
            s.append("\n")
            s.append(self.indent(n))
            s.append(self.process(item, h, n - 2))
            s.append(",")
        h.pop()
        return "".join(s)

    def unwrap(self, d, item):
        """Translate (unwrap) using an optional wrapper function."""
        try:
            md = d.__metadata__
            pmd = getattr(md, "__print__", None)
            if pmd is None:
                return item
            wrappers = getattr(pmd, "wrappers", {})
            fn = wrappers.get(item[0], lambda x: x)
            return (item[0], fn(item[1]))
        except Exception:
            pass
        return item

    def exclude(self, d, item):
        """Check metadata for excluded items."""
        try:
            md = d.__metadata__
            pmd = getattr(md, "__print__", None)
            if pmd is None:
                return False
            excludes = getattr(pmd, "excludes", [])
            return item[0] in excludes
        except Exception:
            pass
        return False



# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Service proxy implementation providing access to web services.

"""

import suds
from suds import *
import suds.bindings.binding
from suds.builder import Builder
import suds.cache
import suds.metrics as metrics
from suds.options import Options
from suds.plugin import PluginContainer
from suds.properties import Unskin
from suds.reader import DefinitionsReader
from suds.resolver import PathResolver
from suds.sax.document import Document
import suds.sax.parser
from suds.servicedefinition import ServiceDefinition
import suds.transport
import suds.transport.https
from suds.umx.basic import Basic as UmxBasic
from suds.wsdl import Definitions
from . import sudsobject

from http.cookiejar import CookieJar
from copy import deepcopy
import http.client

from logging import getLogger
log = getLogger(__name__)


class Client(UnicodeMixin):
    """
    A lightweight web service client.

    @ivar wsdl: The WSDL object.
    @type wsdl:L{Definitions}
    @ivar service: The service proxy used to invoke operations.
    @type service: L{Service}
    @ivar factory: The factory used to create objects.
    @type factory: L{Factory}
    @ivar sd: The service definition
    @type sd: L{ServiceDefinition}
    @ivar messages: The last sent/received messages.
    @type messages: str[2]

    """

    @classmethod
    def items(cls, sobject):
        """
        Extract I{items} from a suds object.

        Much like the items() method works on I{dict}.

        @param sobject: A suds object
        @type sobject: L{Object}
        @return: A list of items contained in I{sobject}.
        @rtype: [(key, value),...]

        """
        return sudsobject.items(sobject)

    @classmethod
    def dict(cls, sobject):
        """
        Convert a sudsobject into a dictionary.

        @param sobject: A suds object
        @type sobject: L{Object}
        @return: A dictionary of items contained in I{sobject}.
        @rtype: dict

        """
        return sudsobject.asdict(sobject)

    @classmethod
    def metadata(cls, sobject):
        """
        Extract the metadata from a suds object.

        @param sobject: A suds object
        @type sobject: L{Object}
        @return: The object's metadata
        @rtype: L{sudsobject.Metadata}

        """
        return sobject.__metadata__

    def __init__(self, url, **kwargs):
        """
        @param url: The URL for the WSDL.
        @type url: str
        @param kwargs: keyword arguments.
        @see: L{Options}

        """
        options = Options()
        options.transport = suds.transport.https.HttpAuthenticated()
        self.options = options
        if "cache" not in kwargs:
            kwargs["cache"] = suds.cache.ObjectCache(days=1)
        self.set_options(**kwargs)
        reader = DefinitionsReader(options, Definitions)
        self.wsdl = reader.open(url)
        plugins = PluginContainer(options.plugins)
        plugins.init.initialized(wsdl=self.wsdl)
        self.factory = Factory(self.wsdl)
        self.service = ServiceSelector(self, self.wsdl.services)
        self.sd = []
        for s in self.wsdl.services:
            sd = ServiceDefinition(self.wsdl, s)
            self.sd.append(sd)
        self.messages = dict(tx=None, rx=None)

    def set_options(self, **kwargs):
        """
        Set options.

        @param kwargs: keyword arguments.
        @see: L{Options}

        """
        p = Unskin(self.options)
        p.update(kwargs)

    def add_prefix(self, prefix, uri):
        """
        Add I{static} mapping of an XML namespace prefix to a namespace.

        Useful for cases when a WSDL and referenced XSD schemas make heavy use
        of namespaces and those namespaces are subject to change.

        @param prefix: An XML namespace prefix.
        @type prefix: str
        @param uri: An XML namespace URI.
        @type uri: str
        @raise Exception: prefix already mapped.

        """
        root = self.wsdl.root
        mapped = root.resolvePrefix(prefix, None)
        if mapped is None:
            root.addPrefix(prefix, uri)
            return
        if mapped[1] != uri:
            raise Exception('"%s" already mapped as "%s"' % (prefix, mapped))

    def last_sent(self):
        """
        Get last sent I{soap} message.
        @return: The last sent I{soap} message.
        @rtype: L{Document}
        """
        return self.messages.get('tx')

    def last_received(self):
        """
        Get last received I{soap} message.
        @return: The last received I{soap} message.
        @rtype: L{Document}
        """
        return self.messages.get('rx')

    def clone(self):
        """
        Get a shallow clone of this object.

        The clone only shares the WSDL. All other attributes are unique to the
        cloned object including options.

        @return: A shallow clone.
        @rtype: L{Client}

        """
        class Uninitialized(Client):
            def __init__(self):
                pass
        clone = Uninitialized()
        clone.options = Options()
        cp = Unskin(clone.options)
        mp = Unskin(self.options)
        cp.update(deepcopy(mp))
        clone.wsdl = self.wsdl
        clone.factory = self.factory
        clone.service = ServiceSelector(clone, self.wsdl.services)
        clone.sd = self.sd
        clone.messages = dict(tx=None, rx=None)
        return clone

    def __unicode__(self):
        s = ["\n"]
        s.append("Suds ( https://fedorahosted.org/suds/ )")
        s.append("  version: %s" % (suds.__version__,))
        if suds.__build__:
            s.append("  build: %s" % (suds.__build__,))
        for sd in self.sd:
            s.append("\n\n%s" % (str(sd),))
        return "".join(s)


class Factory:
    """
    A factory for instantiating types defined in the WSDL.

    @ivar resolver: A schema type resolver.
    @type resolver: L{PathResolver}
    @ivar builder: A schema object builder.
    @type builder: L{Builder}

    """

    def __init__(self, wsdl):
        """
        @param wsdl: A schema object.
        @type wsdl: L{wsdl.Definitions}

        """
        self.wsdl = wsdl
        self.resolver = PathResolver(wsdl)
        self.builder = Builder(self.resolver)

    def create(self, name):
        """
        Create a WSDL type by name.

        @param name: The name of a type defined in the WSDL.
        @type name: str
        @return: The requested object.
        @rtype: L{Object}

        """
        timer = metrics.Timer()
        timer.start()
        type = self.resolver.find(name)
        if type is None:
            raise TypeNotFound(name)
        if type.enum():
            result = sudsobject.Factory.object(name)
            for e, a in type.children():
                setattr(result, e.name, e.name)
        else:
            try:
                result = self.builder.build(type)
            except Exception as e:
                log.error("create '%s' failed", name, exc_info=True)
                raise BuildError(name, e)
        timer.stop()
        metrics.log.debug("%s created: %s", name, timer)
        return result

    def separator(self, ps):
        """
        Set the path separator.

        @param ps: The new path separator.
        @type ps: char

        """
        self.resolver = PathResolver(self.wsdl, ps)


class ServiceSelector:
    """
    The B{service} selector is used to select a web service.

    Most WSDLs only define a single service in which case access by subscript
    is passed through to a L{PortSelector}. This is also the behavior when a
    I{default} service has been specified. In cases where multiple services
    have been defined and no default has been specified, the service is found
    by name (or index) and a L{PortSelector} for the service is returned. In
    all cases, attribute access is forwarded to the L{PortSelector} for either
    the I{first} service or the I{default} service (when specified).

    @ivar __client: A suds client.
    @type __client: L{Client}
    @ivar __services: A list of I{WSDL} services.
    @type __services: list

    """
    def __init__(self, client, services):
        """
        @param client: A suds client.
        @type client: L{Client}
        @param services: A list of I{WSDL} services.
        @type services: list

        """
        self.__client = client
        self.__services = services

    def __getattr__(self, name):
        """
        Attribute access is forwarded to the L{PortSelector}.

        Uses the I{default} service if specified or the I{first} service
        otherwise.

        @param name: Method name.
        @type name: str
        @return: A L{PortSelector}.
        @rtype: L{PortSelector}.

        """
        default = self.__ds()
        if default is None:
            port = self.__find(0)
        else:
            port = default
        return getattr(port, name)

    def __getitem__(self, name):
        """
        Provides I{service} selection by name (string) or index (integer).

        In cases where only a single service is defined or a I{default} has
        been specified, the request is forwarded to the L{PortSelector}.

        @param name: The name (or index) of a service.
        @type name: int|str
        @return: A L{PortSelector} for the specified service.
        @rtype: L{PortSelector}.

        """
        if len(self.__services) == 1:
            port = self.__find(0)
            return port[name]
        default = self.__ds()
        if default is not None:
            port = default
            return port[name]
        return self.__find(name)

    def __find(self, name):
        """
        Find a I{service} by name (string) or index (integer).

        @param name: The name (or index) of a service.
        @type name: int|str
        @return: A L{PortSelector} for the found service.
        @rtype: L{PortSelector}.

        """
        service = None
        if not self.__services:
            raise Exception("No services defined")
        if isinstance(name, int):
            try:
                service = self.__services[name]
                name = service.name
            except IndexError:
                raise ServiceNotFound("at [%d]" % (name,))
        else:
            for s in self.__services:
                if name == s.name:
                    service = s
                    break
        if service is None:
            raise ServiceNotFound(name)
        return PortSelector(self.__client, service.ports, name)

    def __ds(self):
        """
        Get the I{default} service if defined in the I{options}.

        @return: A L{PortSelector} for the I{default} service.
        @rtype: L{PortSelector}.

        """
        ds = self.__client.options.service
        if ds is not None:
            return self.__find(ds)


class PortSelector:
    """
    The B{port} selector is used to select a I{web service} B{port}.

    In cases where multiple ports have been defined and no default has been
    specified, the port is found by name (or index) and a L{MethodSelector} for
    the port is returned. In all cases, attribute access is forwarded to the
    L{MethodSelector} for either the I{first} port or the I{default} port (when
    specified).

    @ivar __client: A suds client.
    @type __client: L{Client}
    @ivar __ports: A list of I{service} ports.
    @type __ports: list
    @ivar __qn: The I{qualified} name of the port (used for logging).
    @type __qn: str

    """
    def __init__(self, client, ports, qn):
        """
        @param client: A suds client.
        @type client: L{Client}
        @param ports: A list of I{service} ports.
        @type ports: list
        @param qn: The name of the service.
        @type qn: str

        """
        self.__client = client
        self.__ports = ports
        self.__qn = qn

    def __getattr__(self, name):
        """
        Attribute access is forwarded to the L{MethodSelector}.

        Uses the I{default} port when specified or the I{first} port otherwise.

        @param name: The name of a method.
        @type name: str
        @return: A L{MethodSelector}.
        @rtype: L{MethodSelector}.

        """
        default = self.__dp()
        if default is None:
            m = self.__find(0)
        else:
            m = default
        return getattr(m, name)

    def __getitem__(self, name):
        """
        Provides I{port} selection by name (string) or index (integer).

        In cases where only a single port is defined or a I{default} has been
        specified, the request is forwarded to the L{MethodSelector}.

        @param name: The name (or index) of a port.
        @type name: int|str
        @return: A L{MethodSelector} for the specified port.
        @rtype: L{MethodSelector}.

        """
        default = self.__dp()
        if default is None:
            return self.__find(name)
        return default

    def __find(self, name):
        """
        Find a I{port} by name (string) or index (integer).

        @param name: The name (or index) of a port.
        @type name: int|str
        @return: A L{MethodSelector} for the found port.
        @rtype: L{MethodSelector}.

        """
        port = None
        if not self.__ports:
            raise Exception("No ports defined: %s" % (self.__qn,))
        if isinstance(name, int):
            qn = "%s[%d]" % (self.__qn, name)
            try:
                port = self.__ports[name]
            except IndexError:
                raise PortNotFound(qn)
        else:
            qn = ".".join((self.__qn, name))
            for p in self.__ports:
                if name == p.name:
                    port = p
                    break
        if port is None:
            raise PortNotFound(qn)
        qn = ".".join((self.__qn, port.name))
        return MethodSelector(self.__client, port.methods, qn)

    def __dp(self):
        """
        Get the I{default} port if defined in the I{options}.

        @return: A L{MethodSelector} for the I{default} port.
        @rtype: L{MethodSelector}.

        """
        dp = self.__client.options.port
        if dp is not None:
            return self.__find(dp)


class MethodSelector:
    """
    The B{method} selector is used to select a B{method} by name.

    @ivar __client: A suds client.
    @type __client: L{Client}
    @ivar __methods: A dictionary of methods.
    @type __methods: dict
    @ivar __qn: The I{qualified} name of the method (used for logging).
    @type __qn: str

    """
    def __init__(self, client, methods, qn):
        """
        @param client: A suds client.
        @type client: L{Client}
        @param methods: A dictionary of methods.
        @type methods: dict
        @param qn: The I{qualified} name of the port.
        @type qn: str

        """
        self.__client = client
        self.__methods = methods
        self.__qn = qn

    def __getattr__(self, name):
        """
        Get a method by name and return it in an I{execution wrapper}.

        @param name: The name of a method.
        @type name: str
        @return: An I{execution wrapper} for the specified method name.
        @rtype: L{Method}

        """
        return self[name]

    def __getitem__(self, name):
        """
        Get a method by name and return it in an I{execution wrapper}.

        @param name: The name of a method.
        @type name: str
        @return: An I{execution wrapper} for the specified method name.
        @rtype: L{Method}

        """
        m = self.__methods.get(name)
        if m is None:
            qn = ".".join((self.__qn, name))
            raise MethodNotFound(qn)
        return Method(self.__client, m)


class Method:
    """
    The I{method} (namespace) object.

    @ivar client: A client object.
    @type client: L{Client}
    @ivar method: A I{WSDL} method.
    @type I{raw} Method.

    """

    def __init__(self, client, method):
        """
        @param client: A client object.
        @type client: L{Client}
        @param method: A I{raw} method.
        @type I{raw} Method.

        """
        self.client = client
        self.method = method

    def __call__(self, *args, **kwargs):
        """Invoke the method."""
        clientclass = self.clientclass(kwargs)
        client = clientclass(self.client, self.method)
        try:
            return client.invoke(args, kwargs)
        except WebFault as e:
            if self.faults():
                raise
            return http.client.INTERNAL_SERVER_ERROR, e

    def faults(self):
        """Get faults option."""
        return self.client.options.faults

    def clientclass(self, kwargs):
        """Get SOAP client class."""
        if _SimClient.simulation(kwargs):
            return _SimClient
        return _SoapClient


class RequestContext:
    """
    A request context.

    Returned by a suds Client when invoking a web service operation with the
    ``nosend`` enabled. Allows the caller to take care of sending the request
    himself and return back the reply data for further processing.

    @ivar envelope: The SOAP request envelope.
    @type envelope: I{bytes}

    """

    def __init__(self, process_reply, envelope):
        """
        @param process_reply: A callback for processing a user defined reply.
        @type process_reply: I{callable}
        @param envelope: The SOAP request envelope.
        @type envelope: I{bytes}

        """
        self.__process_reply = process_reply
        self.envelope = envelope

    def process_reply(self, reply, status=None, description=None):
        """
        Re-entry for processing a successful reply.

        Depending on how the ``retxml`` option is set, may return the SOAP
        reply XML or process it and return the Python object representing the
        returned value.

        @param reply: The SOAP reply envelope.
        @type reply: I{bytes}
        @param status: The HTTP status code.
        @type status: int
        @param description: Additional status description.
        @type description: I{bytes}
        @return: The invoked web service operation return value.
        @rtype: I{builtin}|I{subclass of} L{Object}|I{bytes}|I{None}

        """
        return self.__process_reply(reply, status, description)


class _SoapClient:
    """
    An internal lightweight SOAP based web service operation client.

    Each instance is constructed for specific web service operation and knows
    how to:
      - Construct a SOAP request for it.
      - Transport a SOAP request for it using a configured transport.
      - Receive a SOAP reply using a configured transport.
      - Process the received SOAP reply.

    Depending on the given suds options, may do all the tasks listed above or
    may stop the process at an earlier point and return some intermediate
    result, e.g. the constructed SOAP request or the raw received SOAP reply.
    See the invoke() method for more detailed information.

    @ivar service: The target method.
    @type service: L{Service}
    @ivar method: A target method.
    @type method: L{Method}
    @ivar options: A dictonary of options.
    @type options: dict
    @ivar cookiejar: A cookie jar.
    @type cookiejar: libcookie.CookieJar

    """

    TIMEOUT_ARGUMENT = "__timeout"

    def __init__(self, client, method):
        """
        @param client: A suds client.
        @type client: L{Client}
        @param method: A target method.
        @type method: L{Method}

        """
        self.client = client
        self.method = method
        self.options = client.options
        self.cookiejar = CookieJar()

    def invoke(self, args, kwargs):
        """
        Invoke a specified web service method.

        Depending on how the ``nosend`` & ``retxml`` options are set, may do
        one of the following:
          * Return a constructed web service operation SOAP request without
            sending it to the web service.
          * Invoke the web service operation and return its SOAP reply XML.
          * Invoke the web service operation, process its results and return
            the Python object representing the returned value.

        When returning a SOAP request, the request is wrapped inside a
        RequestContext object allowing the user to acquire a corresponding SOAP
        reply himself and then pass it back to suds for further processing.

        Constructed request data is automatically processed using registered
        plugins and serialized into a byte-string. Exact request XML formatting
        may be affected by the ``prettyxml`` suds option.

        @param args: A list of args for the method invoked.
        @type args: list|tuple
        @param kwargs: Named (keyword) args for the method invoked.
        @type kwargs: dict
        @return: SOAP request, SOAP reply or a web service return value.
        @rtype: L{RequestContext}|I{builtin}|I{subclass of} L{Object}|I{bytes}|
            I{None}

        """
        timer = metrics.Timer()
        timer.start()
        binding = self.method.binding.input
        timeout = kwargs.pop(_SoapClient.TIMEOUT_ARGUMENT, None)
        soapenv = binding.get_message(self.method, args, kwargs)
        timer.stop()
        method_name = self.method.name
        metrics.log.debug("message for '%s' created: %s", method_name, timer)
        timer.start()
        result = self.send(soapenv, timeout=timeout)
        timer.stop()
        metrics.log.debug("method '%s' invoked: %s", method_name, timer)
        return result

    def send(self, soapenv, timeout=None):
        """
        Send SOAP message.

        Depending on how the ``nosend`` & ``retxml`` options are set, may do
        one of the following:
          * Return a constructed web service operation request without sending
            it to the web service.
          * Invoke the web service operation and return its SOAP reply XML.
          * Invoke the web service operation, process its results and return
            the Python object representing the returned value.

        @param soapenv: A SOAP envelope to send.
        @type soapenv: L{Document}
        @return: SOAP request, SOAP reply or a web service return value.
        @rtype: L{RequestContext}|I{builtin}|I{subclass of} L{Object}|I{bytes}|
            I{None}

        """
        location = self.__location()
        log.debug("sending to (%s)\nmessage:\n%s", location, soapenv)
        self.last_sent(soapenv)
        plugins = PluginContainer(self.options.plugins)
        plugins.message.marshalled(envelope=soapenv.root())
        if self.options.prettyxml:
            soapenv = soapenv.str()
        else:
            soapenv = soapenv.plain()
        soapenv = soapenv.encode("utf-8")
        ctx = plugins.message.sending(envelope=soapenv)
        soapenv = ctx.envelope
        if self.options.nosend:
            return RequestContext(self.process_reply, soapenv)
        request = suds.transport.Request(location, soapenv, timeout)
        request.headers = self.__headers()
        try:
            timer = metrics.Timer()
            timer.start()
            reply = self.options.transport.send(request)
            timer.stop()
            metrics.log.debug("waited %s on server reply", timer)
        except suds.transport.TransportError as e:
            content = e.fp and e.fp.read() or ""
            return self.process_reply(content, e.httpcode, tostr(e))
        return self.process_reply(reply.message, None, None)

    def process_reply(self, reply, status, description):
        """
        Process a web service operation SOAP reply.

        Depending on how the ``retxml`` option is set, may return the SOAP
        reply XML or process it and return the Python object representing the
        returned value.

        @param reply: The SOAP reply envelope.
        @type reply: I{bytes}
        @param status: The HTTP status code (None indicates httplib.OK).
        @type status: int|I{None}
        @param description: Additional status description.
        @type description: str
        @return: The invoked web service operation return value.
        @rtype: I{builtin}|I{subclass of} L{Object}|I{bytes}|I{None}

        """
        if status is None:
            status = http.client.OK
        debug_message = "Reply HTTP status - %d" % (status,)
        if status in (http.client.ACCEPTED, http.client.NO_CONTENT):
            log.debug(debug_message)
            return
        #TODO: Consider whether and how to allow plugins to handle error,
        # httplib.ACCEPTED & httplib.NO_CONTENT replies as well as successful
        # ones.
        if status == http.client.OK:
            log.debug("%s\n%s", debug_message, reply)
        else:
            log.debug("%s - %s\n%s", debug_message, description, reply)

        plugins = PluginContainer(self.options.plugins)
        ctx = plugins.message.received(reply=reply)
        reply = ctx.reply

        # SOAP standard states that SOAP errors must be accompanied by HTTP
        # status code 500 - internal server error:
        #
        # From SOAP 1.1 specification:
        #   In case of a SOAP error while processing the request, the SOAP HTTP
        # server MUST issue an HTTP 500 "Internal Server Error" response and
        # include a SOAP message in the response containing a SOAP Fault
        # element (see section 4.4) indicating the SOAP processing error.
        #
        # From WS-I Basic profile:
        #   An INSTANCE MUST use a "500 Internal Server Error" HTTP status code
        # if the response message is a SOAP Fault.
        replyroot = None
        if status in (http.client.OK, http.client.INTERNAL_SERVER_ERROR):
            replyroot = _parse(reply)
            if len(reply) > 0:
                self.last_received(replyroot)
            plugins.message.parsed(reply=replyroot)
            fault = self.__get_fault(replyroot)
            if fault:
                if status != http.client.INTERNAL_SERVER_ERROR:
                    log.warning("Web service reported a SOAP processing fault "
                        "using an unexpected HTTP status code %d. Reporting "
                        "as an internal server error.", status)
                if self.options.faults:
                    raise WebFault(fault, replyroot)
                return http.client.INTERNAL_SERVER_ERROR, fault
        if status != http.client.OK:
            if self.options.faults:
                #TODO: Use a more specific exception class here.
                raise Exception((status, description))
            return status, description

        if self.options.retxml:
            return reply

        result = replyroot and self.method.binding.output.get_reply(
            self.method, replyroot)
        ctx = plugins.message.unmarshalled(reply=result)
        result = ctx.reply
        if self.options.faults:
            return result
        return http.client.OK, result

    def __get_fault(self, replyroot):
        """
        Extract fault information from a SOAP reply.

        Returns an I{unmarshalled} fault L{Object} or None in case the given
        XML document does not contain a SOAP <Fault> element.

        @param replyroot: A SOAP reply message root XML element or None.
        @type replyroot: L{Element}|I{None}
        @return: A fault object.
        @rtype: L{Object}

        """
        def get_fault(envns):
            soapenv = replyroot and replyroot.getChild("Envelope", envns)
            soapbody = soapenv and soapenv.getChild("Body", envns)
            return soapbody and soapbody.getChild("Fault", envns)

        fault = get_fault(suds.bindings.binding.envns) or get_fault(suds.bindings.binding.envns12)
        return fault is not None and UmxBasic().process(fault)

    def __headers(self):
        """
        Get HTTP headers for a HTTP/HTTPS SOAP request.

        @return: A dictionary of header/values.
        @rtype: dict

        """
        action = self.method.soap.action
        if isinstance(action, str):
            action = action.encode("utf-8")
        result = {
            "Content-Type": "text/xml; charset=utf-8",
            "SOAPAction": action}
        result.update(**self.options.headers)
        log.debug("headers = %s", result)
        return result

    def __location(self):
        """Returns the SOAP request's target location URL."""
        return Unskin(self.options).get("location", self.method.location)

    def last_sent(self, d=None):
        """
        Get or set last SOAP sent messages document

        To get the last sent document call the function without parameter.
        To set the last sent message, pass the document as parameter.

        @param d: A SOAP reply dict message key
        @type string: I{bytes}
        @return: The last sent I{soap} message.
        @rtype: L{Document}

        """
        key = 'tx'
        messages = self.client.messages
        if d is None:
            return messages.get(key)
        else:
            messages[key] = d

    def last_received(self, d=None):
        """
        Get or set last SOAP received messages document

        To get the last received document call the function without parameter.
        To set the last sent message, pass the document as parameter.

        @param d: A SOAP reply dict message key
        @type string: I{bytes}
        @return: The last received I{soap} message.
        @rtype: L{Document}

        """
        key = 'rx'
        messages = self.client.messages
        if d is None:
            return messages.get(key)
        else:
            messages[key] = d


class _SimClient(_SoapClient):
    """
    Loopback _SoapClient used for SOAP request/reply simulation.

    Used when a web service operation is invoked with injected SOAP request or
    reply data.

    """

    __injkey = "__inject"

    @classmethod
    def simulation(cls, kwargs):
        """Get whether injected data has been specified in I{kwargs}."""
        return _SimClient.__injkey in kwargs

    def invoke(self, args, kwargs):
        """
        Invoke a specified web service method.

        Uses an injected SOAP request/response instead of a regularly
        constructed/received one.

        Depending on how the ``nosend`` & ``retxml`` options are set, may do
        one of the following:
          * Return a constructed web service operation request without sending
            it to the web service.
          * Invoke the web service operation and return its SOAP reply XML.
          * Invoke the web service operation, process its results and return
            the Python object representing the returned value.

        @param args: Positional arguments for the method invoked.
        @type args: list|tuple
        @param kwargs: Keyword arguments for the method invoked.
        @type kwargs: dict
        @return: SOAP request, SOAP reply or a web service return value.
        @rtype: L{RequestContext}|I{builtin}|I{subclass of} L{Object}|I{bytes}|
            I{None}

        """
        simulation = kwargs.pop(self.__injkey)
        msg = simulation.get("msg")
        if msg is not None:
            assert msg.__class__ is suds.byte_str_class
            return self.send(_parse(msg))
        msg = self.method.binding.input.get_message(self.method, args, kwargs)
        log.debug("inject (simulated) send message:\n%s", msg)
        reply = simulation.get("reply")
        if reply is not None:
            assert reply.__class__ is suds.byte_str_class
            status = simulation.get("status")
            description = simulation.get("description")
            if description is None:
                description = "injected reply"
            return self.process_reply(reply, status, description)
        raise Exception("reply or msg injection parameter expected")


def _parse(string):
    """
    Parses given XML document content.

    Returns the resulting root XML element node or None if the given XML
    content is empty.

    @param string: XML document content to parse.
    @type string: I{bytes}
    @return: Resulting root XML element node or None.
    @rtype: L{Element}|I{None}

    """
    if string:
        return suds.sax.parser.Parser().parse(string=string)
